#!/usr/bin/env ruby

require 'rubygems'
require 'uri'
require 'bio'
require 'json'
require 'securerandom'

# [TODO] integrate this into BioRuby
module Bio
  class GenBank
    def dblink
      fetch('DBLINK')
    end

    def project
      dblink[/Project: (\d+)/, 1]
    end

    def bioproject
      dblink[/BioProject: (\S+)/, 1]
    end

    def biosample
      dblink[/BioSample: (\S+)/, 1]
    end
  end
end

###
### Utilities for RDF generation
###

module RDFSupport
  def new_uuid(prefix = "http://purl.jp/bio/10/genome/uuid/")
    #return "<#{prefix}#{SecureRandom.uuid}>"
    #return "genome:uuid-#{SecureRandom.uuid}"
    return "<urn:uuid:#{SecureRandom.uuid}>"
  end

  def quote(str)
    return str.gsub('\\', '\\\\').gsub("\t", '\\t').gsub("\n", '\\n').gsub("\r", '\\r').gsub('"', '\\"').inspect
  end

  def triple(s, p, o)
    return [s, p, o].join("\t") + " ."
  end

  def default_prefix
    return [
      triple("@prefix", "rdf:", "<http://www.w3.org/1999/02/22-rdf-syntax-ns#>"),
      triple("@prefix", "rdfs:", "<http://www.w3.org/2000/01/rdf-schema#>"),
      #triple("@prefix", "dcterms:", "<http://purl.org/dc/terms/>"),
      triple("@prefix", "xsd:", "<http://www.w3.org/2001/XMLSchema#>"),
      triple("@prefix", "skos:", "<http://www.w3.org/2004/02/skos/core#>"),
      triple("@prefix", "sio:", "<http://semanticscience.org/resource/>"),
      #triple("@prefix", "so:", "<http://purl.org/obo/owl/SO#>"),
      triple("@prefix", "obo:", "<http://purl.obolibrary.org/obo/>"),
      triple("@prefix", "faldo:", "<http://biohackathon.org/resource/faldo#>"),
    ]
  end

  def usdate2date(str)
    return Date.parse(str).strftime("%Y-%m-%d")  
  end
end

###
### Mapping INSDC or RefSeq db_xref to Identifiers.org
###

# https://gist.github.com/3985701
# https://gist.github.com/4146256
class RS_ID
  include RDFSupport

  def initialize
    @rs_id = JSON.parse(File.read(File.dirname(__FILE__) + "/rs_id.json"))
  end

  def fetch(db)
    @rs_id[db]
  end

  def labels
    @rs_id.sort.each do |db, hash|
      puts triple("insdc:#{hash['class']}", "rdfs:label", quote(hash['label']))
    end
  end
end

###
### Mapping INSDC or RefSeq feature table to Sequence Ontology
###

# https://gist.github.com/3650401
class FT_SO
  def initialize
    @data = JSON.parse(File.read(File.dirname(__FILE__) + "/ft_so.json"))
  end

  # ftso = FT_SO.new
  # puts ftso.so_id("-10_signal")  # => "SO:0000175"
  def so_id(feature)
    if hash = @data[feature]
      return hash["so_id"]
    end
  end

  # puts ftso.so_id("-10_signal")  # => "obo:SO_0000175"
  def obo_id(feature)
    return "obo:" + so_id(feature).sub(':', '_')
  end

  # puts ftso.so_term("-10_signal")  # => "minus_10_signal"
  def so_term(feature)
    if hash = @data[feature]
      return hash["so_term"]
    end
  end

  # puts ftso.so_desc("-10_signal")  # => "A conserved region about 10-bp upstream of ..."
  def so_desc(feature)
    if hash = @data[feature]
      return hash["so_desc"]
    end
  end

  # puts ftso.ft_desc("-10_signal")  # => "Pribnow box; a conserved region about 10 bp upstream of ..."
  def ft_desc(feature)
    if hash = @data[feature]
      return hash["ft_desc"]
    end
  end
end

###
### Convert INSDC or RefSeq entries to RDF
###

class INSDC2RDF

  include RDFSupport

  def initialize(io = nil, opts = nil)
    @datasource = opts[:datasource]
    @seqtype = opts[:seqtype]
    @rs_id = RS_ID.new
    @ft_so = FT_SO.new

    # @datasource can be 'insdc' or 'refseq'
    @entry_prefix = "http://identifiers.org/#{@datasource}/"

    @gene = {}
    @xref_warn = {}
    @feature_count = Hash.new(0)

    set_prefixes

    puts prefix
    puts

    parse_entry(io) if io
  end

  attr_accessor :prefix

  def self.prefixes
    puts prefix
  end

  def set_prefixes
    @prefix = default_prefix + [
      #triple("@prefix", "genome:", "<http://purl.jp/bio/10/genome/>"),
      #triple("@prefix", "insdc:", "<http://insdc.org/owl/>"),
      triple("@prefix", "insdc:", "<http://ddbj.nig.ac.jp/ontologies/sequence/>"),
    ]
  end

  def xref(subject, db, id)
    case db
    when "GI"
      id = "GI:#{id}"
    when "HOMD"
      id.sub!(/^tax_/, '')
    when "ECOCYC"
      #id = "ECOCYC:#{id}"
    when "ERIC", "HMP", "PSEUDO", "Pathema"
      unless @xref_warn[db]
        $stderr.puts "Warning: Need to register '#{db}' in Identifiers.org"
        @xref_warn[db] = true
      end
    end

    if hash = @rs_id.fetch(db)
      uri = "<#{hash['prefix']}/#{id}>"
      puts triple(subject, "rdfs:seeAlso", uri)
      puts triple(uri, "rdfs:label", quote(id))
      puts triple(uri, "sio:SIO_000068", "<#{hash['prefix']}>") + "  # sio:is-part-of"
    else
      unless @xref_warn[db]
        $stderr.puts "Error: New database '#{db}' found. Add it to the rs_id.json file and/or Identifiers.org."
        @xref_warn[db] = true
      end
    end
  end

  ###
  ### FALDO http://biohackathon.org/faldo
  ###

  def new_feature_uri(feature, from, to, strand, count = false)
    if count
      "<#{@sequence_id}#feature:#{from}-#{to}:#{strand}/#{feature}:#{count}>"
    else
      "<#{@sequence_id}#feature:#{from}-#{to}:#{strand}/#{feature}>"
    end
  end

  def new_region_uri(from, to, strand)
    "<#{@sequence_id}#region:#{from}-#{to}:#{strand}>"
  end

  def new_position_uri(pos, strand)
    "<#{@sequence_id}#position:#{pos}:#{strand}>"
  end

  def new_location(object, pos)
    puts triple(object, "insdc:location", quote(pos))
    locations = Bio::Locations.new(pos)

    # [TODO] annotation located over the origin of a circular genome can be faldo:begin > faldo:end even in ForwardStrandPosition
    # e.g., join(800..900,1000..1024,1..234) will be faldo:begin 800 and faldo:end 234
    min, max = locations.span
    strand = locations.first.strand
    if strand > 0
      fuzzy_min = locations.first.lt
      fuzzy_max = locations.last.gt
      strand_min = locations.first.strand
      strand_max = locations.last.strand
    else
      fuzzy_min = locations.last.lt
      fuzzy_max = locations.first.gt
      strand_min = locations.last.strand
      strand_max = locations.first.strand
    end
    region_id = new_faldo_region(object, min, max, strand_min, strand_max, fuzzy_min, fuzzy_max)

    return region_id, locations
  end

  def add_subparts(locations, feature_type)
    count = 1
    subparts = []
    locations.each do |loc|
      subpart_id = new_uuid
      subfeature_id = new_feature_uri(feature_type[:term], loc.from, loc.to, loc.strand)

      #puts triple(subpart_id, "obo:so_part_of", region_id)
      puts triple(subpart_id, "sio:SIO_000300", count) + "  # sio:has-value"
      puts triple(subpart_id, "sio:SIO_000628", subfeature_id) + "  # sio:referes-to"

      puts triple(subfeature_id, "rdf:type", feature_type[:id]) + "  # SO:#{feature_type[:term]}"
      new_faldo_region(subfeature_id, loc.from, loc.to, loc.strand, loc.strand, loc.lt, loc.gt)
      subparts << subpart_id
      count += 1
    end
    return subparts
  end

  def new_faldo_region(object, min, max, strand_min, strand_max, fuzzy_min = nil, fuzzy_max = nil)
    region_id = new_region_uri(min, max, strand_min)
    region_min = new_position_uri(min, strand_min)
    region_max = new_position_uri(max, strand_max)

    puts triple(object, "faldo:location", region_id)
    puts triple(region_id, "rdf:type", "faldo:Region")
    if strand_min > 0
      puts triple(region_id, "faldo:begin", region_min)
      puts triple(region_id, "faldo:end", region_max)
      new_faldo_position(region_min, min, "faldo:ForwardStrandPosition", fuzzy_min)
      new_faldo_position(region_max, max, "faldo:ForwardStrandPosition", fuzzy_max)
    else
      puts triple(region_id, "faldo:begin", region_max)
      puts triple(region_id, "faldo:end", region_min)
      new_faldo_position(region_min, min, "faldo:ReverseStrandPosition", fuzzy_min)
      new_faldo_position(region_max, max, "faldo:ReverseStrandPosition", fuzzy_max)
    end

    return region_id
  end

  def new_faldo_position(pos_id, pos, strand, fuzzy = nil)
    puts triple(pos_id, "faldo:position", pos)
    puts triple(pos_id, "faldo:reference", @sequence_uri)
    puts triple(pos_id, "rdf:type", strand)
    if fuzzy
      puts triple(pos_id, "rdf:type", "faldo:FuzzyPosition")
    else
      puts triple(pos_id, "rdf:type", "faldo:ExactPosition")
    end
  end

  ###
  ### Main
  ###

  # [TODO] rewrite parse_entry and subsequent methods to use Bio::Sequence for EMBL support
  def parse_entry(io)
    # Read INSDC or RefSeq entry
    Bio::FlatFile.auto(io).each do |entry|
      @entry = entry
      @features = entry.features
      @source = @features.shift
      parse_sequence
      parse_source
      parse_genes
      parse_features
    end
  end

  ###
  ### Sequence
  ###

  # [TODO]
  # * bind multiple sequences by BioProject ID?
  # * add a flag for sequencing status representing complete/draft?
  def parse_sequence
    @sequence_id = "#{@entry_prefix}#{@entry.acc_version}"
    @sequence_uri = "<#{@sequence_id}>"

    # [TODO] how to automatically identify the input is chromosome/plasmid/contig/...?
    sequence_type(@seqtype)
    # [TODO] obtain rdfs:label from source /chromosome (eukaryotes) /plasmid (prokaryotes) -> see insdc:source_chromosome, insdc:source_plasmid
    sequence_label(@entry.definition)
    sequence_version(@entry.acc_version)
    sequence_length(@entry.nalen)
    # [TODO] find appropriate REST URI to retreive genomic DNA sequence by <@sequence_id.fasta>
    sequence_seq(@entry.acc_version)
    sequence_form(@entry.circular)
    # [TODO] sequenced date: updated date in the source DB or generation date of the RDF data?
    sequence_date(@entry.date)
    # [TODO] rdfs:seeAlso (like UniProt) or dc:relation, owl:sameAs
    sequence_link_gi(@entry.gi.sub('GI:',''))
    sequence_link_accver(@entry.acc_version)
    if bioproject = @entry.bioproject
      sequence_link_bioproject(bioproject)
    elsif project = @entry.project
      sequence_link_bioproject("PRJNA#{project}")
    end
    if biosample = @entry.biosample
      sequence_link_biosample(biosample)
    end
    # [TODO] how to deal with direct submissions (references without PMID)?
    sequence_ref(@entry.references)
  end

  def sequence_type(so = "SO:sequence")
    case so
    when /0000001/, "SO:region", "SO:sequence"
      puts triple(@sequence_uri, "rdf:type", "obo:SO_0000001") + "  # SO:sequence"
    when /0000340/, "SO:chromosome"
      puts triple(@sequence_uri, "rdf:type", "obo:SO_0000340") + "  # SO:chromosome"
    when /0000155/, "SO:plasmid"
      puts triple(@sequence_uri, "rdf:type", "obo:SO_0000155") + "  # SO:plasmid"
    when /0000736/, "SO:organelle_sequence"
      puts triple(@sequence_uri, "rdf:type", "obo:SO_0000736") + "  # SO:organelle_sequence"
    when /0000819/, "SO:mitochondrial_chromosome"
      puts triple(@sequence_uri, "rdf:type", "obo:SO_0000819") + "  # SO:mitochondrial_chromosome"
    when /0000740/, "SO:plastid_sequence"
      puts triple(@sequence_uri, "rdf:type", "obo:SO_0000740") + "  # SO:plastid_sequence"
    when /0000719/, "SO:ultracontig"
      puts triple(@sequence_uri, "rdf:type", "obo:SO_0000719") + "  # SO:ultracontig"
    when /0000148/, "SO:supercontig", "SO:scaffold"
      puts triple(@sequence_uri, "rdf:type", "obo:SO_0000148") + "  # SO:supercontig/scaffold"
    when /0000149/, "SO:contig"
      puts triple(@sequence_uri, "rdf:type", "obo:SO_0000149") + "  # SO:contig"
    else
      puts triple(@sequence_uri, "rdf:type", "obo:SO_0000353") + "  # SO:sequence_assembly"
    end
  end

  def sequence_label(str)
    # Use "name:" key in the JSON representation
    puts triple(@sequence_uri, "rdfs:label", quote(str))
  end

  def sequence_version(str)
    puts triple(@sequence_uri, "insdc:sequence_version", quote(str))
  end

  def sequence_length(int)
    puts triple(@sequence_uri, "insdc:sequence_length", int)
  end

  def sequence_seq(str)
    # [TODO] where to obtain the actual DNA sequence? in what format?
    #fasta_uri = "<http://togows.org/entry/nucleotide/#{str}.fasta>"
    fasta_uri = "<http://www.ncbi.nlm.nih.gov/nuccore/#{str}?report=fasta>"
    puts triple(@sequence_uri, "insdc:sequence_fasta", fasta_uri)
  end

  def sequence_form(form)
    case form
    when "linear"
      puts triple(@sequence_uri, "rdf:type", "obo:SO_0000987") + "  # SO:linear"
    when "circular"
      puts triple(@sequence_uri, "rdf:type", "obo:SO_0000988") + "  # SO:circular"
    end
  end

  def sequence_date(date)
    puts triple(@sequence_uri, "insdc:sequence_date", quote(usdate2date(date))+"^^xsd:date")
  end

  def sequence_link_gi(str)
    xref(@sequence_uri, 'GI', str)
  end

  def sequence_link_accver(str, source_db = 'RefSeq')
    # [TODO] automatically distinguish RefSeq/GenBank/DDBJ entries by the prefix of accession IDs?
    # [TODO] register GenBank/DDBJ in rs_id.json to enable the above
    xref(@sequence_uri, source_db, str)
  end

  def sequence_link_bioproject(str)
    id_pfx = "http://identifiers.org/bioproject"
    xref_id = "<#{id_pfx}/#{str}>"
    puts triple(@sequence_uri, 'insdc:dblink', xref_id)
    puts triple(xref_id, 'rdfs:label', quote(str))
    puts triple(xref_id, 'rdf:type', "<#{id_pfx}#Entry>")
  end

  def sequence_link_biosample(str)
    id_pfx = "http://identifiers.org/biosample"
    xref_id = "<#{id_pfx}/#{str}>"
    xred_type = "<#{id_pfx}#Entry>"
    puts triple(@sequence_uri, 'insdc:dblink', xref_id)
    puts triple(xref_id, 'rdfs:label', quote(str))
    puts triple(xred_id, 'rdf:type', xref_type)
  end

  def sequence_ref(refs)
    refs.each do |ref|
      if pmid = ref.pubmed
        if pmid.length > 0
          xref(@sequence_uri, 'PubMed', pmid)
        end
      end
    end
  end

  ###
  ### Source
  ###

  def parse_source
    # Use @sequence_uri for @source_uri
    @source_uri = @sequence_uri

    hash = @source.to_hash
    source_location(@source.position)
    source_link(hash["db_xref"])
    hash.delete("db_xref")
    source_qualifiers(hash)
  end

  def source_location(pos)
    new_location(@source_uri, pos)
  end

  def source_link(links)
    links.each do |link|
      db, entry_id = link.split(':', 2)
      xref(@source_uri, db, entry_id)
    end
  end

  def source_qualifiers(hash)
    hash.each do |qual, vals|
      vals.each do |val|
        if val == true
          puts triple(@source_uri, "insdc:#{qual}", true)
        else        
          data = val.to_s.gsub(/\s+/, ' ').strip
          if data[/^\d+$/]
            puts triple(@source_uri, "insdc:#{qual}", data)
          else
            puts triple(@source_uri, "insdc:#{qual}", quote(data))
          end
        end
      end
    end
  end

  ###
  ### Genes
  ###

  # parse genes first to collect gene IDs for mRNA, CDS etc. (then parse rest of features)
  def parse_genes
    genes = @features.select {|x| x.feature == "gene"}
  
    genes.each do |gene|
      @feature_count[gene.feature] += 1
      locations = Bio::Locations.new(gene.position)
      gene_id = new_feature_uri("gene", locations.first.from, locations.last.to, locations.first.strand, @feature_count[gene.feature])
      hash = gene.to_hash

      # try to cache gene IDs in the @gene hash for linking with other features (CDS, mRNA etc.)
      if hash["locus_tag"]
        locus_tag = hash["locus_tag"].first
        @gene[locus_tag] = gene_id
      elsif hash["gene"]
        gene = hash["gene"].first
        @gene[gene] = gene_id
      else
        # [TODO] where else to find gene name?
      end
    end
  end

  ###
  ### Features (part of gene: CDS, mRNA, misc_RNA, precursor_RNA, ncRNA, tRNA, rRNA)
  ###

  def parse_features
    @features.each do |feat|
      feature = feat.feature
      position = feat.position

      # try to link gene-related features (CDS, mRNA etc.) by matching /locus_tag or /gene qualifier values
      hash = feat.to_hash
      gene_id = locus_tag = gene = nil
      if hash["locus_tag"]
        if locus_tag = hash["locus_tag"].first
          gene_id = @gene[locus_tag]
        end
      elsif hash["gene"]
        if gene = hash["gene"].first
          gene_id = @gene[gene]
        end
      end

      # re-use URI for genes otherwise generate new URI
      if feature == "gene"
        feature_id = gene_id
      else
        @feature_count[feature] += 1
        locations = Bio::Locations.new(position)
        min, max = locations.span
        strand = locations.first.strand
        feature_id = new_feature_uri(feature, min, max, strand, @feature_count[feature])
      end

      # add type by Sequence Ontology
      so_id = "SO:0000001"
      so_obo_id = "obo:SO_0000001"
      so_term = "region"
      if so_id = @ft_so.so_id(feature)
        if so_id != "undefined"
          so_obo_id = @ft_so.obo_id(feature)
          so_term = @ft_so.so_term(feature)
        end
      end

      # feature types and labels
      puts triple(feature_id, "rdf:type", so_obo_id) + "  # SO:#{so_term}"
      puts triple(feature_id, "rdfs:label", quote(locus_tag || gene || feature))
      if locus_tag || gene
        puts triple(feature_id, "skos:prefLabel", quote(locus_tag || gene))
        if hash["gene_synonym"]
          hash["gene_synonym"].first.split(/;\s+/).each do |synonym|
            puts triple(feature_id, "skos:altLabel", quote(synonym))
          end
        end
      end

      # feature qualifiers
      parse_qualifiers(feature_id, hash)

      # parent-child relationship (gene -> mRNA|CDS|misc_RNA etc.)
      parent_uri = @sequence_uri
      if gene_id and gene_id != feature_id
        parent_uri = gene_id
        puts triple(gene_id, "sio:SIO_010080", feature_id) + "  # sio:is-transcribed-into"
      end
      puts triple(feature_id, "obo:so_part_of", parent_uri)

      # add FALDO location and subparts (exons etc.)
      region_id, locations = new_location(feature_id, position)
      if locations.count > 1
        if gene_id
          # link to exons in join(exon1, exon2, ...)
          feature_type = { :id => "obo:SO_0000147", :term => "exon" }
        else
          # [TODO] need to confirm that if there are any features having subparts other than exons
          feature_type = { :id => "obo:SO_0000001", :term => "region" }
        end
        subparts = add_subparts(locations, feature_type)
        #puts triple(feature_id, "obo:so_has_part", "(#{subparts.join(' ')})")  # rdf:List
        puts triple(feature_id, "sio:SIO_000974", subparts.join(', ')) + "  # sio:has-ordered-part"
      end
    end
    $stderr.puts "Features: #{@feature_count.to_json}"
  end

  def parse_qualifiers(feature_id, hash)
    hash.each do |qual, vals|
      vals.each do |val|
        if val == true
          puts triple(feature_id, "insdc:#{qual}", true)
        else
          data = val.to_s.gsub(/\s+/, ' ').strip
          case qual
          when "protein_id"
            xref(feature_id, 'Protein', val)
          when "db_xref"
            db, id = val.split(':', 2)
            # ad hoc
            if db == 'InterPro' and @entry.acc_version[/(NC_010994.1|NC_014958.1|NC_015385.1|NC_015386.1|NC_015387.1|NC_015388.1|NC_015389.1)/]
              # PRJNA59115/NC_010994.1
              # PRJNA62225/NC_014958.1
              # PRJNA65781/NC_015385.1
              # PRJNA65781/plasmids/NC_015386.1
              # PRJNA65783/NC_015387.1
              # PRJNA65785/NC_015388.1
              # PRJNA65787/NC_015389.1
              #       /db_xref="InterPro:Chromosomal replication control,
              #       initiator (DnaA)/regulator (Hda"
              xref(feature_id, db, id) if id[/IPR\d+/]
            elsif db == "ASAP" and @entry.acc_version[/(NC_017263.1|NC_017264.1|NC_017265.1|NC_017266.1)/]
              # PRJNA158537/plasmids/NC_017263.1
              # PRJNA158537/plasmids/NC_017264.1
              # PRJNA158537/NC_017265.1
              # PRJNA158537/plasmids/NC_017266.1
              #       /db_xref="ASAP:BBE-0004740"
              #       /db_xref="ASAP:BBE-0004740 ERIC"
              xref(feature_id, db, id) unless id[/\s/]
            elsif db == 'TIGRFAM' and @entry.acc_version[/NC_013418.2/]
              # PRJNA41287/NC_013418.2
              #       /db_xref="TIGRFAM:TIGR00336, TIGR01740"
              #       /db_xref="TIGRFAM:TIGR00197; TF"
              if id[/,/]
                ids = id.split(',').map{|x| x.strip}
                ids.each do |x|
                  xref(feature_id, db, x)
                end
              else
                xref(feature_id, db, id.sub(/;.*/, ''))
              end
            else
              xref(feature_id, db, id)
            end
          else
            if data[/^\d+$/]
              puts triple(feature_id, "insdc:#{qual}", data)
            else
              puts triple(feature_id, "insdc:#{qual}", quote(data))
            end
          end
        end
      end
    end
  end

end


if __FILE__ == $0
  require 'getoptlong'

  args = GetoptLong.new(
    [ '--datasource', '-d', GetoptLong::REQUIRED_ARGUMENT ],
    [ '--seqtype', '-t', GetoptLong::REQUIRED_ARGUMENT ],
    [ '--prefixes', '-p', GetoptLong::NO_ARGUMENT ],
  )

  opts = {
    :seqtype => "SO:sequence",
    :datasource => "insdc",
  }

  args.each_option do |name, value|
    case name
    when /--datasource/
      opts[:datasource] = value
    when /--seqtype/
      opts[:seqtype] = value
    when /--prefixes/
      opts[:prefixes] = true
    end
  end

  if opts[:prefixes]
    INSDC2RDF.new
  else
    INSDC2RDF.new(ARGF, opts)
  end
end

__END__
